/**
 * Copyright (c) 2015, STARSCHEMA LTD.
 * All rights reserved.

 * Redistribution and use in source and binary forms, with or without modification, are permitted
 * provided that the following conditions are met:

 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Modified by
 * @author  Data Virtuality GmbH
 * @date    $Date$
 */

package net.starschema.clouddb.jdbc;

import com.google.api.client.auth.oauth2.Credential;
import com.google.api.services.bigquery.BigqueryScopes;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Method;
import java.net.URLDecoder;
import java.security.GeneralSecurityException;
import java.sql.Array;
import java.sql.Blob;
import java.sql.CallableStatement;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.NClob;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLClientInfoException;
import java.sql.SQLException;
import java.sql.SQLFeatureNotSupportedException;
import java.sql.SQLWarning;
import java.sql.SQLXML;
import java.sql.Savepoint;
import java.sql.Statement;
import java.sql.Struct;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.Executor;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.starschema.clouddb.cmdlineverification.Oauth2Bigquery;
import net.starschema.clouddb.jdbc.list.CallContainer;

import org.apache.log4j.Logger;

import com.google.api.services.bigquery.Bigquery;

// import net.starschema.clouddb.bqjdbc.logging.Logger;

/**
 * The connection class which builds the connection between BigQuery and the
 * Driver
 *
 * @author Gunics Balázs, Horváth Attila
 *
 */
public class BQConnection implements Connection {
    /** Variable to store auto commit mode */
    private boolean autoCommitEnabled = false;

    /** Instance log4j.Logger */
    Logger logger;
    /**
     * The bigquery client to access the service.
     */
    private Bigquery bigquery = null;

    /**
     * The projectid which needed for the queries.
     */
    private String projectId = null;
    /** Boolean to determine if the Connection is closed */
    private boolean isclosed = false;

    /** Boolean to determine, to use or doesn't use the ANTLR parser */
    private boolean transformQuery = false;

    /** getter for transformQuery */
    public boolean getTransformQuery() {
        return transformQuery;
    }

    /** List to contain sql warnings in */
    private List<SQLWarning> SQLWarningList = new ArrayList<SQLWarning>();

    /** String to contain the url except the url prefix */
    private String URLPART = null;

	// EXTENSION
	// ISSUE: DVCORE-2295
	// FIXED BY: Roman Pedchenko
	// DATE: 2014.10.12
	private boolean allowLargeResults = false;
	// END EXTENSION

	// EXTENSION
	// ISSUE: DVCORE-2506
	// FIXED BY: Roman Pedchenko
	// DATE: 2014.11.20
	private String largeResultsDefaultDataset = null;
	// END EXTENSION

	// EXTENSION
	// ISSUE: DVCORE-2472
	// FIXED BY: Roman Pedchenko
	// DATE: 2014.10.30
	private long tableMaxResults = 10000l;
	// END EXTENSION

	// EXTENSION
	// ISSUE: DVCORE-2462
	// FIXED BY: Roman Pedchenko
	// DATE: 2014.11.19
	private CallContainer callContainer = new CallContainer();
	// END EXTENSION

	private Integer fetchSize = null;

	/**
	 * Data source name
	 */
	private String datasourceName;

	/**
	 * Google authorization code.
	 */
	private String authCode;

	/**
	 * Redirect URL. This must match the one associated with your client ID.
	 * Note: the redirectUri needs to match between auth and token requests.
	 */
	private String redirectUri;

	private String authType;

	private boolean serviceAccount;

	private String userId;

	private String userKey;

	private String credentialFactory;

	public interface OAuth2 {
		Credential authorize(String userId, String userKey, String authCode, String redirectUri, String datasourceName, String... scopes) throws Exception;
	}

	/**
     * Extracts the JDBC URL then makes a connection to the Bigquery.
     *
     * @param serverdata
     *            the JDBC connection URL
     * @param loginProp
     *
     * @throws GeneralSecurityException
     * @throws IOException
     * @throws SQLException
     */
    public BQConnection(String url, Properties loginProp) throws SQLException {

        this.logger = Logger.getLogger(this.getClass());
        this.URLPART = url;
        this.isclosed = false;

        //If the URL contains user/password we'll use that
        //If it doesn't we'll look them in the loginProp
        boolean containUserPassword = false;

        String projectid;

        if (url.contains("&user=") && url.contains("&password=")) {
            containUserPassword = true;
            this.logger.debug("url contains &user and &password");
        } else this.logger.debug("url doesn't contains &user and &password");

        //getting the user/password for the connection
        if (containUserPassword) {
            //getting the User/Password from the URL
            int passwordindex = url.indexOf("&password=");
            int userindex = url.indexOf("&user=");
            try {
                userId = URLDecoder.decode(url.substring(
                        userindex + "&user=".length(), passwordindex), "UTF-8");
                userKey = URLDecoder.decode(
                        url.substring(passwordindex + "&password=".length()),
                        "UTF-8");
            } catch (UnsupportedEncodingException e2) {
                throw new BQSQLException(e2);
            }
        } else {
            //getting the User/Password from property
            userId = loginProp.getProperty("user");
            userKey = loginProp.getProperty("password");
        }

        //getting the project ID
        try {
            projectid = URLDecoder.decode(
                    url.substring(url.lastIndexOf(":") + 1), "UTF-8");
            this.logger.debug("projectid + end of url: " + projectid);
            //we either got parameters with ?:
            if (projectid.contains("?")) {
            this.projectId = projectid.substring(0, projectid.indexOf("?"));
            }
            //or we got the projectID right
            else this.projectId = projectid;
        } catch (UnsupportedEncodingException e1) {
            throw new BQSQLException(e1);
        }
        //lets replace the : with __ and . with _
        // EXTENSION
        // ISSUE: DVCORE-2295
        // FIXED BY: Roman Pedchenko
        // DATE: 2014.10.12
        // projectId = projectId.replace(":" , "__").replace(".", "_");
        projectId = BQSupportFuncts.encodeId(projectId);
        // END EXTENSION
        //Connect with serviceAccount?
        String lowerCasedUrl = url.toLowerCase();
        /*if (lowerCasedUrl.contains("?withserviceaccount=true")) {
            serviceAccount = true;
            this.logger.debug("url contains ?withServiceAccount=true");
        } else {
            String sa = "";
            if (loginProp.getProperty("withServiceAccount") != null) sa = loginProp.getProperty("withServiceAccount");
            if (loginProp.getProperty("withServiceaccount") != null) sa = loginProp.getProperty("withServiceaccount");
            if (loginProp.getProperty("withserviceAccount") != null) sa = loginProp.getProperty("withserviceAccount");
            if (loginProp.getProperty("withserviceaccount") != null) sa = loginProp.getProperty("withserviceaccount");
            if (loginProp.getProperty("WithServiceAccount") != null) sa = loginProp.getProperty("WithServiceAccount");
            if (loginProp.getProperty("WithServiceaccount") != null) sa = loginProp.getProperty("WithServiceaccount");
            if (loginProp.getProperty("WithserviceAccount") != null) sa = loginProp.getProperty("WithserviceAccount");
            if (loginProp.getProperty("Withserviceaccount") != null) sa = loginProp.getProperty("Withserviceaccount");
            serviceAccount = false;
           serviceAccount = Boolean.parseBoolean(sa);
            logger.debug("from the properties we got for withServiceAccount the following: " +
                    sa + " which converts to: " + Boolean.toString(serviceAccount));
        }*/

        //do we want to transform Queries?
        if (lowerCasedUrl.contains("transformquery=true")) {
            this.transformQuery = true;
            this.logger.debug("url contains transformQuery=true");
        } else {
            String lp = "";
            if (loginProp.getProperty("transformQuery") != null) lp = loginProp.getProperty("transformQuery");
            if (loginProp.getProperty("TransformQuery") != null) lp = loginProp.getProperty("TransformQuery");
            if (loginProp.getProperty("Transformquery") != null) lp = loginProp.getProperty("Transformquery");
            if (loginProp.getProperty("transformquery") != null) lp = loginProp.getProperty("transformquery");
            this.transformQuery = false;
            this.transformQuery = Boolean.parseBoolean(lp);
            logger.debug("from the properties we got for transformQuery the following: " +
                    lp + " which converts to: " + Boolean.toString(transformQuery));
        }

	    Matcher mtch = Pattern.compile("[\\?\\&]authType=([^\\?&]*)", Pattern.CASE_INSENSITIVE).matcher(url);
	    if (mtch.find() && mtch.groupCount() == 1) {
		    authType = mtch.group(1);
	    }
	    mtch = Pattern.compile("[\\?\\&]credentialFactory=([^\\?&]*)", Pattern.CASE_INSENSITIVE).matcher(url);
	    if (mtch.find() && mtch.groupCount() == 1) {
		    credentialFactory = mtch.group(1);
	    }
		// EXTENSION
		// ISSUE: DVCORE-2295
		// FIXED BY: Roman Pedchenko
		// DATE: 2014.10.12
		mtch = Pattern.compile("[\\?\\&]allowLargeResults=([^\\?&]*)", Pattern.CASE_INSENSITIVE).matcher(url);
		if (mtch.find() && mtch.groupCount() == 1) {
			allowLargeResults = Boolean.valueOf(mtch.group(1));
		}
		// END EXTENSION
		// EXTENSION
		// ISSUE: DVCORE-2506
		// FIXED BY: Roman Pedchenko
		// DATE: 2014.11.20
		mtch = Pattern.compile("[\\?\\&]largeResultsDefaultDataset=([^\\?&]*)", Pattern.CASE_INSENSITIVE).matcher(url);
		if (mtch.find() && mtch.groupCount() == 1) {
			largeResultsDefaultDataset = mtch.group(1);
		}
		// END EXTENSION
		// EXTENSION
		// ISSUE: DVCORE-2472
		// FIXED BY: Roman Pedchenko
		// DATE: 2014.10.30
		mtch = Pattern.compile("[\\?\\&]tableMaxResults=([^\\?&]*)", Pattern.CASE_INSENSITIVE).matcher(url);
		if (mtch.find() && mtch.groupCount() == 1) {
			tableMaxResults = Long.valueOf(mtch.group(1));
		}
		// END EXTENSION
		mtch = Pattern.compile("[\\?\\&]fetchSize=([^\\?&]*)", Pattern.CASE_INSENSITIVE).matcher(url);
		if (mtch.find() && mtch.groupCount() == 1) {
			fetchSize = Integer.valueOf(mtch.group(1));
		}
	    mtch = Pattern.compile("[\\?\\&]datasourceName=([^\\?&]*)", Pattern.CASE_INSENSITIVE).matcher(url);
	    if (mtch.find() && mtch.groupCount() == 1) {
		    datasourceName = mtch.group(1);
	    }
	    mtch = Pattern.compile("[\\?\\&]authCode=([^\\?&]*)", Pattern.CASE_INSENSITIVE).matcher(url);
	    if (mtch.find() && mtch.groupCount() == 1) {
		    authCode = mtch.group(1);
	    }
	    mtch = Pattern.compile("[\\?\\&]redirectUri=([^\\?&]*)", Pattern.CASE_INSENSITIVE).matcher(url);
	    if (mtch.find() && mtch.groupCount() == 1) {
		    try {
		        redirectUri = URLDecoder.decode(mtch.group(1), "UTF-8");
		    } catch (UnsupportedEncodingException e) {
			    throw new BQSQLException(e);
		    }
	    }

	    if (authType != null) {
		    if (!authType.equalsIgnoreCase(Oauth2Bigquery.AuthorizationType.SERVICE.toString()) && !authType.equalsIgnoreCase(Oauth2Bigquery.AuthorizationType.OAUTH.toString()))
			    throw new BQSQLException("Google BigQuery authType property has an incorrect value. Please provide a correct one from the following list: OAUTH, SERVICE.");
	    }
	    serviceAccount = (authType == null || (authType != null && authType.equalsIgnoreCase(Oauth2Bigquery.AuthorizationType.SERVICE.toString())));

	    /**
	     * Lets make a connection:
	     *
	     */
	    //do we have a serviceaccount to connect with?
	    if (serviceAccount) {
		    if (userId == null || userId.isEmpty())
			    throw new BQSQLException("Google BigQuery connector userId property is not set. The BigQuery data source won't be created.");
		    if (userKey == null || userKey.isEmpty())
			    throw new BQSQLException("Google Analytics connector userKey property is not set. The BigQuery data source won't be created.");
		    try {
			    this.bigquery = Oauth2Bigquery.authorizeviaservice(userId, userKey);
			    this.logger.info("Authorized with service account");
		    } catch (GeneralSecurityException e) {
			    throw new BQSQLException(e);
		    } catch (IOException e) {
			    throw new BQSQLException(e);
		    }
	    } else {
			this.logger.info("Authorizing with Oauth");
			if (credentialFactory == null) {
				// default behavior
				this.bigquery = Oauth2Bigquery.authorizeviainstalled(userId, userKey, getScopes());
			} else {
				try {
					Class<?> clazz = Class.forName(credentialFactory);
					Method method = clazz.getMethod("authorize", String.class, String.class, String.class, String.class, String.class, String[].class);
					Credential credential = (Credential)method.invoke(null, userId, userKey, authCode, redirectUri, datasourceName, getScopes().toArray(new String[0]));
					this.bigquery = Oauth2Bigquery.authorizeByCredential(credential);
				} catch (Exception e) {
					throw new BQSQLException(e);
				}
			}
			this.logger.info("Authorized with Oauth");
		}
	    logger.debug("The project id for this connections is: " + this.projectId);
    }

	public void authorizeByCredential(OAuth2 oAuth2) throws Exception {
		logger.debug("Authorizing as installed app.");
		if (this.bigquery == null && oAuth2 != null) {
			logger.debug("Creating a new bigquery client.");
			this.bigquery = Oauth2Bigquery.authorizeByCredential(oAuth2.authorize(userId, userKey, authCode, redirectUri, datasourceName, getScopes().toArray(new String[0])));
			this.logger.info("Authorized with OAuth");
		}
	}

	private List<String> getScopes() {
		List<String> Scopes = new ArrayList<String>();
		Scopes.add(BigqueryScopes.BIGQUERY);
		return Scopes;
	}
    
	// EXTENSION
	// ISSUE: DVCORE-2295
	// FIXED BY: Roman Pedchenko
	// DATE: 2014.10.12
	public boolean getAllowLargeResults() {
		return allowLargeResults;
	}

	public void setAllowLargeResults(boolean allowLargeResults) {
		this.allowLargeResults = allowLargeResults;
	}
	// END EXTENSION
	// EXTENSION
	// ISSUE: DVCORE-2506
	// FIXED BY: Roman Pedchenko
	// DATE: 2014.11.20
	public String getLargeResultsDefaultDataset() {
		return largeResultsDefaultDataset;
	}

	public void setLargeResultsDefaultDataset(String largeResultsDefaultDataset) {
		this.largeResultsDefaultDataset = largeResultsDefaultDataset;
	}
	// END EXTENSION
	// EXTENSION
	// ISSUE: DVCORE-2472
	// FIXED BY: Roman Pedchenko
	// DATE: 2014.10.30
    public long getTableMaxResults() {
		return tableMaxResults;
	}

	public void setTableMaxResults(long tableMaxResults) {
		this.tableMaxResults = tableMaxResults;
	}
	// END EXTENSION

	public Integer getFetchSize() {
		return fetchSize;
	}

	public void setFetchSize(Integer fetchSize) {
		this.fetchSize = fetchSize;
	}

	public String getDatasourceName() {
		return datasourceName;
	}

	public void setDatasourceName(String datasourceName) {
		this.datasourceName = datasourceName;
	}

	public String getAuthCode() {
		return authCode;
	}

	public void setAuthCode(String authCode) {
		this.authCode = authCode;
	}

	public String getRedirectUri() {
		return redirectUri;
	}

	public void setRedirectUri(String redirectUri) {
		this.redirectUri = redirectUri;
	}

	public String getAuthType() {
		return authType;
	}

	public void setAuthType(String authType) {
		this.authType = authType;
	}

	public boolean isServiceAccount() {
		return serviceAccount;
	}

	/**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Uses SQLWarningList.clear() to clear all warnings
     * </p>
     */
    @Override
    public void clearWarnings() throws SQLException {
        if (this.isclosed) {
            throw new BQSQLException("Connection is closed.");
        }
        this.SQLWarningList.clear();
    }

	/**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Sets bigquery to null and isclosed to true if the connection is not
     * already closed else no operation is performed
     * </p>
     */
    @Override
    public void close() throws SQLException {
        if (!this.isclosed) {
            this.bigquery = null;
            this.isclosed = true;
        }
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Throws Exception
     * </p>
     *
     * @throws SQLException
     *             <p>
     *             There is no Commit in Google BigQuery + Connection Status
     *             </p>
     */
    @Override
    public void commit() throws SQLException {
        if (this.isclosed) {
            throw new BQSQLException(
                    "There's no commit in Google BigQuery.\nConnection Status: Closed.");
        }/*
        else {
            throw new BQSQLException(
                    "There's no commit in Google BigQuery.\nConnection Status: Open.");
        }*/
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public Array createArrayOf(String typeName, Object[] elements)
            throws SQLException {
        throw new BQSQLException("Not implemented."
                + "createArrayOf(String typeName, Object[] elements)");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public Blob createBlob() throws SQLException {
        throw new BQSQLException("Not implemented." + "createBlob()");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public Clob createClob() throws SQLException {
        throw new BQSQLException("Not implemented." + "createClob()");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public NClob createNClob() throws SQLException {
        throw new BQSQLException("Not implemented." + "createNClob()");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public SQLXML createSQLXML() throws SQLException {
        throw new BQSQLException("Not implemented." + "createSQLXML()");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Creates a new BQStatement object with the projectid in this Connection
     * </p>
     *
     * @return a new BQStatement object with the projectid in this Connection
     * @throws SQLException
     *             if the Connection is closed
     */
    @Override
    public Statement createStatement() throws SQLException {
        if (this.isclosed) {
            throw new BQSQLException("Connection is closed.");
        }
        logger.debug("Creating statement with resultsettype: forward only," +
        		" concurrency: read only");
        return new BQStatement(this.projectId, this);
    }

    /** {@inheritDoc} */
    @Override
    public Statement createStatement(int resultSetType, int resultSetConcurrency)
            throws SQLException {
        if (this.isClosed()) {
            throw new BQSQLException("The Connection is Closed");
        }
        logger.debug("Creating statement with resultsettype: " + resultSetType
                + " concurrency: " + resultSetConcurrency);
        return new BQStatement(this.projectId, this, resultSetType,
                resultSetConcurrency);
    }

    /** {@inheritDoc} */
    @Override
    public Statement createStatement(int resultSetType,
            int resultSetConcurrency, int resultSetHoldability)
            throws SQLException {
        throw new BQSQLException("Not implemented."
                + "createStaement(int,int,int)");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public Struct createStruct(String typeName, Object[] attributes)
            throws SQLException {
        throw new BQSQLException("Not implemented."
                + "createStruct(string,object[])");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Getter for autoCommitEnabled
     * </p>
     *
     * @return auto commit state;
     */
    @Override
    public boolean getAutoCommit() throws SQLException {
        return this.autoCommitEnabled;
    }

    /**
     * Getter method for the authorized bigquery client
     */
    public Bigquery getBigquery() {
        return this.bigquery;
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Return projectid
     * </p>
     *
     * @return projectid Contained in this Connection instance
     */
    @Override
    public String getCatalog() throws SQLException {
        logger.debug("function call getCatalog returning projectId: " + projectId);
        return this.projectId;
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public Properties getClientInfo() throws SQLException {
        throw new BQSQLException("Not implemented." + "getClientInfo()");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public String getClientInfo(String name) throws SQLException {
        throw new BQSQLException("Not implemented." + "");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * There's no commit.
     * </p>
     *
     * @return CLOSE_CURSORS_AT_COMMIT
     */
    @Override
    public int getHoldability() throws SQLException {
        return ResultSet.CLOSE_CURSORS_AT_COMMIT;
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Return a new BQDatabaseMetadata object constructed from this Connection
     * instance
     * </p>
     *
     * @return a new BQDatabaseMetadata object constructed from this Connection
     *         instance
     */
    @Override
    public DatabaseMetaData getMetaData() throws SQLException {
        BQDatabaseMetadata metadata = new BQDatabaseMetadata(this);
        return metadata;
    }

    /**
     * Getter method for projectid
     */
    public String getProjectId() {
        return this.projectId;
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Transactions are not supported.
     * </p>
     *
     * @return TRANSACTION_NONE
     */
    @Override
    public int getTransactionIsolation() throws SQLException {
        return java.sql.Connection.TRANSACTION_NONE;
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public Map<String, Class<?>> getTypeMap() throws SQLException {
        throw new BQSQLException("Not implemented." + "getTypeMap()");
    }

    /**
     *
     * @return The URL which is in the JDBC drivers connection URL
     */
    public String getURLPART() {
        return this.URLPART;
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * If SQLWarningList is empty returns null else it returns the first item
     * Contained inside <br>
     * Subsequent warnings will be chained to this SQLWarning.
     * </p>
     *
     * @return SQLWarning (The First item Contained in SQLWarningList) + all
     *         others chained to it
     */
    @Override
    public SQLWarning getWarnings() throws SQLException {
        if (this.isclosed) {
            throw new BQSQLException("Connection is closed.");
        }
        if (this.SQLWarningList.isEmpty()) {
            return null;
        }

        SQLWarning forreturn = this.SQLWarningList.get(0);
        this.SQLWarningList.remove(0);
        if (!this.SQLWarningList.isEmpty()) {
            for (SQLWarning warning : this.SQLWarningList) {
                forreturn.setNextWarning(warning);
            }
        }
        return forreturn;
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * returns the status of isclosed boolean
     * </p>
     */
    @Override
    public boolean isClosed() throws SQLException {
        return this.isclosed;
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * The driver is read only at this stage.
     * </p>
     *
     * @return true
     */
    @Override
    public boolean isReadOnly() throws SQLException {
        return true;
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Sends a query to BigQuery to get all the datasets contained in the
     * project accociated with this Connection object and checks if it's
     * succeeded
     * </p>
     *
     * @throws SQLException
     */
    @Override
    public boolean isValid(int timeout) throws SQLException {
        if (this.isclosed) {
            return false;
        }
        if (timeout < 0) {
            throw new BQSQLException(
                    "Timeout value can't be negative. ie. it must be 0 or above; timeout value is: "
                            + String.valueOf(timeout));
        }
        try {
            // EXTENSION
            // ISSUE: DVCORE-2295
            // FIXED BY: Roman Pedchenko
            // DATE: 2014.10.12
            // this.bigquery.datasets().list(this.projectId.replace("__", ":").replace("_", ".")).execute();
            this.bigquery.datasets().list(BQSupportFuncts.decodeId(this.projectId)).execute();
            // END EXTENSION
        }
        catch (IOException e) {
            return false;
        }
        return true;
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * We returns the original sql statement
     * </p>
     *
     * @return sql - the original statement
     */
    @Override
    public String nativeSQL(String sql) throws SQLException {
        logger.debug("Function called nativeSQL() " + sql);
        return sql;
        // TODO
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public CallableStatement prepareCall(String sql) throws SQLException {
        throw new BQSQLException("Not implemented." + "prepareCall(string)");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public CallableStatement prepareCall(String sql, int resultSetType,
            int resultSetConcurrency) throws SQLException {
        throw new BQSQLException("Not implemented."
                + "prepareCall(String,int,int)");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public CallableStatement prepareCall(String sql, int resultSetType,
            int resultSetConcurrency, int resultSetHoldability)
            throws SQLException {
        throw new BQSQLException("Not implemented."
                + "prepareCall(string,int,int,int)");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Creates and returns a PreparedStatement object
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public PreparedStatement prepareStatement(String sql) throws SQLException {
        this.logger.debug("Creating Prepared Statement project id is: "
                + this.projectId + " with parameters:");
        this.logger.debug(sql);
        PreparedStatement stm = new BQPreparedStatement(sql, this.projectId,
                this);
        return stm;
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public PreparedStatement prepareStatement(String sql, int autoGeneratedKeys)
            throws SQLException {
        throw new BQSQLException("Not implemented."
                + "prepareStatement(string,int)");
    }

    /** {@inheritDoc} */
    @Override
    public PreparedStatement prepareStatement(String sql, int resultSetType,
            int resultSetConcurrency) throws SQLException {
        this.logger.debug("Creating Prepared Statement" +
                " project id is: " + this.projectId +
                ", resultSetType (int) is: " + String.valueOf(resultSetType) +
            ", resultSetConcurrency (int) is: " + String.valueOf(resultSetConcurrency)
                + " with parameters:");
        this.logger.debug(sql);
        PreparedStatement stm = new BQPreparedStatement(sql, this.projectId,
                this, resultSetType, resultSetConcurrency);
        return stm;
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public PreparedStatement prepareStatement(String sql, int resultSetType,
            int resultSetConcurrency, int resultSetHoldability)
            throws SQLException {
        throw new BQSQLException("Not implemented."
                + "prepareStatement(String,int,int,int)");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public PreparedStatement prepareStatement(String sql, int[] columnIndexes)
            throws SQLException {
        throw new BQSQLException("Not implemented."
                + "prepareStatement(String,int[])");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public PreparedStatement prepareStatement(String sql, String[] columnNames)
            throws SQLException {
        throw new BQSQLException("Not implemented."
                + "prepareStatement(String,String[])");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public void releaseSavepoint(Savepoint savepoint) throws SQLException {
        throw new BQSQLException("Not implemented."
                + "releaseSavepoint(Savepoint)");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public void rollback() throws SQLException {
        logger.debug("function call: rollback() not implemented ");
        //throw new BQSQLException("Not implemented." + "rollback()");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public void rollback(Savepoint savepoint) throws SQLException {
        throw new BQSQLException("Not implemented." + "rollback(savepoint)");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Setter for autoCommitEnabled
     * </p>
     *
     *
     */
    @Override
    public void setAutoCommit(boolean autoCommit) throws SQLException {
        this.autoCommitEnabled = autoCommit;
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public void setCatalog(String catalog) throws SQLException {
        throw new BQSQLException("Not implemented." + "setCatalog(catalog)");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public void setClientInfo(Properties properties)
            throws SQLClientInfoException {
        SQLClientInfoException e = new SQLClientInfoException();
        e.setNextException(new BQSQLException(
                "Not implemented. setClientInfo(properties)"));
        throw e;
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public void setClientInfo(String name, String value)
            throws SQLClientInfoException {
        SQLClientInfoException e = new SQLClientInfoException();
        e.setNextException(new BQSQLException(
                "Not implemented. setClientInfo(properties)"));
        throw e;
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public void setHoldability(int holdability) throws SQLException {
        if (this.isclosed) {
            throw new BQSQLException("Connection is closed.");
        }
        throw new BQSQLException("Not implemented." + "setHoldability(int)");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * BigQuery is ReadOnly always so this is a noop
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public void setReadOnly(boolean readOnly) throws SQLException {
        if (this.isClosed()) {
            throw new BQSQLException("This Connection is Closed");
        }
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public Savepoint setSavepoint() throws SQLException {
        throw new BQSQLException("Not implemented." + "setSavepoint()");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public Savepoint setSavepoint(String name) throws SQLException {
        throw new BQSQLException("Not implemented." + "setSavepoint(String)");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public void setTransactionIsolation(int level) throws SQLException {
        throw new BQSQLException("Not implemented."
                + "setTransactionIsolation(int)");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Not implemented yet.
     * </p>
     *
     * @throws BQSQLException
     */
    @Override
    public void setTypeMap(Map<String, Class<?>> map) throws SQLException {
        throw new BQSQLException("Not implemented."
                + "setTypeMap(Map<String, Class<?>>");
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Returns false to everything
     * </p>
     *
     * @return false
     */
    @Override
    public boolean isWrapperFor(Class<?> arg0) throws SQLException {
        return arg0.isAssignableFrom(Bigquery.class);
    }

    /**
     * <p>
     * <h1>Implementation Details:</h1><br>
     * Always throws SQLException
     * </p>
     *
     * @return nothing
     * @throws SQLException
     *             Always
     */
    @SuppressWarnings("unchecked")
	@Override
    public <T> T unwrap(Class<T> arg0) throws SQLException {
    	if (arg0.isAssignableFrom(Bigquery.class))
    		return (T) bigquery;
        throw new BQSQLException("Not found");
    }

	/* (non-Javadoc)
     * @see java.sql.Connection#setSchema(java.lang.String)
     */
	@Override
	public void setSchema(String schema) throws SQLException {
		// TODO Auto-generated method stub

	}

	/* (non-Javadoc)
	 * @see java.sql.Connection#getSchema()
	 */
	@Override
	public String getSchema() throws SQLException {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
     * @see java.sql.Connection#abort(java.util.concurrent.Executor)
     */
	@Override
	public void abort(Executor executor) throws SQLException {
		// TODO Auto-generated method stub

	}

	/* (non-Javadoc)
     * @see java.sql.Connection#setNetworkTimeout(java.util.concurrent.Executor, int)
     */
	@Override
	public void setNetworkTimeout(Executor executor, int milliseconds) throws SQLException {
		throw new SQLFeatureNotSupportedException("setNetworkTimeout is not supported");
	}

	/* (non-Javadoc)
	 * @see java.sql.Connection#getNetworkTimeout()
	 */
	@Override
	public int getNetworkTimeout() throws SQLException {
		return 0;
	}

	// EXTENSION
	// ISSUE: DVCORE-2462
	// FIXED BY: Roman Pedchenko
	// DATE: 2014.11.19
	public CallContainer getCallContainer() {
		return callContainer;
	}

	public void setCallContainer(CallContainer callContainer) {
		this.callContainer = callContainer;
	}
	// END EXTENSION
}
