/**
 * Copyright (c) 2015, STARSCHEMA LTD.
 * All rights reserved.

 * Redistribution and use in source and binary forms, with or without modification, are permitted
 * provided that the following conditions are met:

 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * <p>
 * This class implements functions such as connecting to Bigquery, Checking out
 * the results and displaying them on console
 * </p>
 *
 * Modified by
 * @author  Data Virtuality GmbH
 * @date    $Date$
 */

package net.starschema.clouddb.jdbc;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.Properties;
import java.util.TimeZone;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.bind.annotation.adapters.HexBinaryAdapter;

import net.starschema.clouddb.jdbc.list.SourceTable;

import org.apache.log4j.Logger;

import com.google.api.services.bigquery.Bigquery;
import com.google.api.services.bigquery.Bigquery.Jobs.Insert;
import com.google.api.services.bigquery.model.DatasetList.Datasets;
import com.google.api.services.bigquery.model.GetQueryResultsResponse;
import com.google.api.services.bigquery.model.Job;
import com.google.api.services.bigquery.model.JobConfiguration;
import com.google.api.services.bigquery.model.JobConfigurationQuery;
import com.google.api.services.bigquery.model.ProjectList.Projects;
import com.google.api.services.bigquery.model.ProjectReference;
import com.google.api.services.bigquery.model.Table;
import com.google.api.services.bigquery.model.TableList.Tables;
import com.google.api.services.bigquery.model.TableReference;
import com.google.api.services.bigquery.model.TableRow;
import com.google.api.services.bigquery.model.TableCell;

// import net.starschema.clouddb.bqjdbc.logging.Logger;

/**
 * This class contains static methods for interacting with BigQuery
 *
 * @author Gunics Balázs, Horváth Attila
 */
public class BQSupportFuncts {
	public static final int NULLTYPE_REQUIRED = 0;
	public static final int NULLTYPE_NULLABLE = 1;
	public static final int NULLTYPE_REPEATED = 2;

	/** log4j.Logger instance */
    // static Logger logger = new Logger(BQSupportFuncts.class.getName());
    static Logger logger = Logger.getLogger(BQSupportFuncts.class.getName());

    /**
     * Constructs a valid BigQuery JDBC driver URL from the specified properties
     * File
     *
     * @param properties Properties File most commonly read by
     *            ReadUrlFromPropFile(String)
     * @return a valid BigQuery JDBC driver URL or null if it fails to load
     * @throws UnsupportedEncodingException
     * @see #ReadUrlFromPropFile(String)
     */
    public static String constructUrlFromPropertiesFile(Properties properties)
            throws UnsupportedEncodingException {
        String ProjectId = properties.getProperty("projectid");
        logger.debug("projectId is: " + ProjectId);
        String User = properties.getProperty("user");
        String Password = properties.getProperty("password");
        String transformQuery = null;

        transformQuery = properties.getProperty("transformquery");

        String forreturn = "";

        if (properties.getProperty("type").equals("installed")) {
            if (User != null && Password != null && ProjectId != null) {
                forreturn = BQDriver.getURLPrefix()
                        + URLEncoder.encode(ProjectId, "UTF-8");
            } else {
                return null;
            }
        } else if (properties.getProperty("type").equals("service")) {
                if (User != null && Password != null && ProjectId != null) {
                    forreturn = BQDriver.getURLPrefix()
                            + URLEncoder.encode(ProjectId, "UTF-8")
                            + "?withServiceAccount=true";
            } else {
                    return null;
                }
        } else {
                return null;
            }

        if (transformQuery != null) {
            return forreturn + "?transformQuery=" + transformQuery;
        } else return forreturn;
        }

    /**
     * Displays results of a Query on the console
     *
     * @param bigquery     A valid authorized Bigquery instance
     * @param projectId    The exact Id of the Project associated with the completed
     *            BigQuery job
     * @param completedJob The Job reference of the completed job
     * @throws IOException <p>
     *             If the request fails to get the QueryResult of the specified
     *             job in the given ProjectId
     *             </p>
     */
    public static void displayQueryResults(Bigquery bigquery, String projectId,
            Job completedJob) throws IOException {
        projectId = BQSupportFuncts.decodeId(projectId);
        GetQueryResultsResponse queryResult = bigquery
                .jobs()
                .getQueryResults(projectId,
                        completedJob.getJobReference().getJobId()).execute();
        List<TableRow> rows = queryResult.getRows();
        System.out.print("\nQuery Results:\n------------\n");
        for (TableRow row : rows) {
            for (TableCell field : row.getF()) {
                System.out.printf("%-20s", field.getV());
            }
            System.out.println();
        }
    }

    /**
     * Return a list of Projects which contains the String catalogname
     *
     * @param catalogName The String which the id of the result Projects must contain
     * @param Connection  A valid BQConnection instance
     * @return a list of Projects which contains the String catalogname
     * @throws IOException <p>
     *             if the initialization of requesting the list of all Projects
     *             (to be sorted from) fails
     *             </p>
     */
    private static List<Projects> getCatalogs(String catalogName,
            BQConnection Connection) throws IOException {
        logger.debug("Function call getCatalogs catalogName: " +
                (catalogName != null ? catalogName : "null"));
        List<Projects> projects = Connection.getBigquery().projects().list()
                .execute().getProjects();

        //since we'll use . -> _ : -> __ conversion, it's easier to
        // replace all the . : before we do any compare
        if (projects != null && projects.size() != 0) {             //we got projects!
            for (Projects projects2 : projects) {
                projects2.setId(projects2.getId().replace(".", "_").replace(":", "__"));
                //updating the reference too
                ProjectReference projRef = projects2.getProjectReference();
                projRef.setProjectId(projRef.getProjectId().replace(".", "_").replace(":", "__"));
                projects2.setProjectReference(projRef);
            }
            if (catalogName != null) {
                List<Projects> ProjectsSearch = new ArrayList<Projects>();
                for (Projects project : projects) {
                    if (project.getId().contains(catalogName) || project.getNumericId().toString().contains(catalogName)) {
                        ProjectsSearch.add(project);
                    }
                }
                if (ProjectsSearch.size() == 0) {
                    return null;
                } else {
                    return ProjectsSearch;
                }
            } else {
                return projects;
            }
        } else {
            return null;
        }
    }

    /**
     * Parses a (instance of dataset).getId() and gives back the id only for the
     * dataset
     *
     * @param getId (instance of dataset).getId()
     * @return the id only for the dataset
     */
    public static String getDatasetIdFromDatasetDotGetId(String getId) {
        /*logger.debug("Function call getDatasetIdFromDatasetDotGetId" + getId +
                "return is: " + getId.substring(getId.lastIndexOf(":") + 1));*/
        return getId.substring(getId.lastIndexOf(":") + 1);
    }

    /**
     * Parses a (instance of table).getid() and gives back the id only for the
     * dataset
     *
     * @param getId (instance of table).getid()
     * @return the id only for the dataset
     */
    public static String getDatasetIdFromTableDotGetId(String getId) {
        /*logger.debug("Function call getDatasetIdFromTableDotGetId" + getId +
                "return is: " + getId.substring(getId.lastIndexOf(":") + 1, getId.lastIndexOf(".")));*/
        return getId.substring(getId.lastIndexOf(":") + 1, getId.lastIndexOf("."));
    }

    /**
     * Returns a list of Datasets, which are associated with the Project which's
     * id is exactly ProjectId, and their name matches datasetnamepattern
     *
     * @param datasetname The String the dataset's id must contain
     * @param projectId   The Id of the Project the dataset is preferably contained in
     * @param connection  A valid BQConnection Instance
     * @return a list of Datasets, which are associated with the Project which's
     *         id is exactly ProjectId, and their name contains datasetname
     * @throws IOException <p>
     *             if the request to get Projects that match the given ProjectId
     *             fails
     *             </p>
     */
    private static List<Datasets> getDatasets(String datasetname,
            String projectId, BQConnection connection) throws IOException {
        logger.debug("function call getDatasets, " +
            "datasetName: " + (datasetname != null ? datasetname : "null") +
            ", projectId: " + (projectId != null ? projectId : "null"));
        projectId = BQSupportFuncts.decodeId(projectId);
        List<Datasets> datasetcontainer = connection.getBigquery().datasets()
                .list(projectId).execute().getDatasets();

        if (datasetcontainer != null && datasetcontainer.size() != 0) {
            if (datasetname != null) {
                List<Datasets> datasetsSearch = new ArrayList<Datasets>();
                for (Datasets in : datasetcontainer) {
                    if (matchPattern(getDatasetIdFromDatasetDotGetId(in.getId()),
                            datasetname)) {
                        datasetsSearch.add(in);
                    }
                }
                if (datasetsSearch.size() == 0) {
                    return null;
                } else {
                    return datasetsSearch;
                }
            } else {
                return datasetcontainer;
            }
        } else {
            return null;
        }
    }

    /**
     * Parses a (instance of dataset)/(instance of table).getid() and gives back
     * the id only for the Project
     *
     * @param getId (instance of dataset)/(instance of table).getid()
     * @returnthe the id only for the Project
     */
    public static String getProjectIdFromAnyGetId(String getId) {
        int pos = getId.indexOf(":");        // The first appearance of ":"
		if (getId.indexOf(":", pos + 1) != -1) {
            // If there's a second ":" we'll use it
         // (there must be a second ":" !!)
            pos = getId.indexOf(":", pos + 1);
        }
        String ret = getId.substring(0, pos); // Cutting out the project id
        return ret;
    }

    /**
     * Returns the result of a completed query
     *
     * @param bigquery     Instance of authorized Bigquery client
     * @param projectId    The id of the Project the completed job was run in
     * @param completedJob The Job instance of the completed job
     * @return the result of a completed query specified by projectId and
     *         completedJob
     * @throws IOException <p>
     *             if the request to get QueryResults specified by the given
     *             ProjectId and Job id fails
     *             </p>
     */
    public static GetQueryResultsResponse getQueryResults(Bigquery bigquery,
            String projectId, Job completedJob) throws IOException {
        GetQueryResultsResponse queryResult = bigquery.jobs()
                .getQueryResults(projectId,
                        completedJob.getJobReference().getJobId()).execute();
        long totalRows = queryResult.getTotalRows().longValue();
        if (totalRows == 0) {
            return queryResult;
        }
        while (totalRows > (long) queryResult.getRows().size()) {
            queryResult.getRows().addAll(
                bigquery.jobs()
                    .getQueryResults(projectId,
                            completedJob.getJobReference().getJobId())
                            .setStartIndex(BigInteger.valueOf((long) queryResult.getRows().size()))
                    .execute()
                            .getRows());
        }
        return queryResult;
    }

    /**
     * Returns the result of a completed query
     *
     * @param bigquery     Instance of authorized Bigquery client
     * @param projectId    The id of the Project the completed job was run in
     * @param completedJob The Job instance of the completed job
     * @return the result of a completed query specified by projectId and
     *         completedJob
     * @throws IOException <p>
     *             if the request to get QueryResults specified by the given
     *             ProjectId and Job id fails
     *             </p>
     */
    public static GetQueryResultsResponse getQueryResultsDivided(Bigquery bigquery,
            String projectId, Job completedJob, BigInteger startAtRow, int fetchCount) throws IOException {
        GetQueryResultsResponse queryResult;
        queryResult = bigquery.jobs()
                .getQueryResults(projectId,
                        completedJob.getJobReference().getJobId())
                        .setStartIndex(startAtRow)
                        .setMaxResults((long) fetchCount).execute();
        return queryResult;
    }

    /**
     * Returns the status of a job
     *
     * @param myjob     Instance of Job
     * @param bigquery  Instance of authorized Bigquery client
     * @param projectId The id of the Project the job is contained in
     * @return the status of the job
     * @throws IOException <p>
     *             if the request to get the job specified by myjob and
     *             projectId fails
     *             </p>
     */
    public static String getQueryState(Job myjob, Bigquery bigquery,
            String projectId) throws IOException {
        projectId = BQSupportFuncts.decodeId(projectId);
        Job pollJob = bigquery.jobs()
                .get(projectId, myjob.getJobReference().getJobId()).execute();
        BQSupportFuncts.logger.info("Job status: "
                + pollJob.getStatus().getState()
                + " ; "
                + pollJob.getJobReference().getJobId()
                + " ; "
                + (System.currentTimeMillis() - pollJob.getStatistics()
                        .getStartTime()));
        return pollJob.getStatus().getState();
    }

    /**
     * Parses a (instance of table).getid() and gives back the id only for the
     * table
     *
     * @param getId (instance of table).getid()
     * @return the id only for the table
     */
    public static String getTableIdFromTableDotGetId(String getId) {
        return getId.substring(getId.lastIndexOf(".") + 1);
    }

    /**
     * Returns a list of Tables which's id matches TablenamePattern and are
     * exactly in the given Project and Dataset
     *
     * @param tablename  String that the tableid must contain
     * @param projectId  The exact Id of the Project that the tables must be in
     * @param datasetId  The exact Id of the Dataset that the tables must be in
     * @param connection Instance of a valid BQConnection
     * @return a list of Tables which's id contains Tablename and are exactly in
     *         the given Project and Dataset
     * @throws IOException <p>
     *             if the request to get all tables (to sort from) specified by
     *             ProjectId, DatasetId fails
     *             </p>
     */
    public static List<Tables> getTables(String tableNamePattern,
            String projectId, String datasetId, BQConnection connection)
            throws IOException {
        logger.debug("Function call getTables : " +
            "tableNamePattern: " + (tableNamePattern != null ? tableNamePattern : "null") +
                ", projectId: " + (projectId != null ? projectId : "null") +
            ", datasetID:" + (datasetId != null ? datasetId : "null") +
            "connection");
        projectId = BQSupportFuncts.decodeId(projectId);
        final com.google.api.services.bigquery.Bigquery.Tables.List tablesList = connection.getBigquery().tables()
                .list(projectId, datasetId);
        if (connection.getTableMaxResults() > 0)
        	tablesList.setMaxResults(connection.getTableMaxResults());
		List<Tables> tables = tablesList.execute().getTables();
        if (tables != null && tables.size() != 0) {
            if (tableNamePattern != null) {
                List<Tables> tablesSearch = new ArrayList<Tables>();
                for (Tables in : tables) {
                    if (matchPattern(getTableIdFromTableDotGetId(in.getId()),
                            tableNamePattern)) {
                        tablesSearch.add(in);
                    }
                }
                if (tablesSearch.size() == 0) {
                    logger.debug("returning null");
                    return null;
                } else {
                    return tablesSearch;
                }
            } else {
                return tables;
            }
        } else {
            logger.debug("returning null");
            return null;
        }
    }

    /**
     * Gets Tables information from specific projects matching catalog,
     * tablenamepattern and datasetidpatterns
     *
     * @param connection Valid instance of BQConnection
     * @return List of Table
     * @throws IOException <p>
     *             if the initialization of requesting the list of all Projects
     *             (to be sorted from) fails<br>
     *             if the request to get Projects that match the given ProjectId
     *             fails<br>
     *             if the request to get all tables (to sort from) specified by
     *             ProjectId, DatasetId fails<br>
     *             if the request to get table information based on ProjectId
     *             DatasetId TableId fails
     *             <p>
     */
    public static List<Table> getTables(BQConnection connection,
            String catalog, String schema, String tablename) throws IOException {
        List<Table> RET = new ArrayList<Table>();
        logger.debug("Function call getTables : " +
                "catalog: " + (catalog != null ? catalog : "null") +
                ", schema: " + (schema != null ? schema : "null") +
                ", tablename:" + (tablename != null ? tablename : "null") +
                "connection");
        //getting the projects for this connection
        List<Projects> Projects = BQSupportFuncts.getCatalogs(catalog,
                connection);

        if (Projects != null && Projects.size() != 0) {
            for (Projects proj : Projects) {
                List<Datasets> datasetlist = null;
                datasetlist = BQSupportFuncts.getDatasets(schema, proj.getId(),
                        connection);
                if (datasetlist != null && datasetlist.size() != 0) {
                    for (Datasets dataset : datasetlist) {
                        List<Tables> tables = null;

                        tables = BQSupportFuncts.getTables(tablename, proj
                                .getId(),
                                BQSupportFuncts
                                        .getDatasetIdFromDatasetDotGetId(dataset
                                                .getId()), connection);

                        if (tables != null && tables.size() != 0) {
                            for (Tables table : tables) {
                                //TODO replace projID __ -> : _ -> .
                                String datasetString = BQSupportFuncts
                                        .getDatasetIdFromDatasetDotGetId(dataset.getId());
                                String tableString = BQSupportFuncts
                                        .getTableIdFromTableDotGetId(table.getId());
                                logger.debug("Calling connection.getBigquery().tables() " +
                                        "dataset is: " + datasetString +
                                        ", table is: " + tableString +
                                        ", project is: " + BQSupportFuncts.decodeId(proj.getId()));
                                try {
									Table tbl = connection.getBigquery().tables()
									        .get(BQSupportFuncts.decodeId(proj.getId()),
									                datasetString,
									                tableString)
									        .execute();
									if (tbl != null) {
									    RET.add(tbl);
									}
								} catch (IOException e) {
									logger.warn("Error getting metadata for BigQuery table: " + table.getTableReference().toPrettyString(), e);
								}
                            }
                        }
                    }
                }
            }
            if (RET.size() == 0) {
                return null;
            } else {
                return RET;
            }
        } else {
            return null;
        }
    }

    /**
     * <br>
     * Some DatabaseMetaData methods take arguments that are String patterns.
     * These arguments all have names such as fooPattern. Within a pattern
     * String, "%" means match any substring of 0 or more characters, and "_"
     * means match any one character. Only metadata entries matching the search
     * pattern are returned. If a search pattern argument is set to null, that
     * argument's criterion will be dropped from the search.
     *
     * Checks if the input String matches the pattern string which may contain
     * %, which means it can be any character
     *
     * @param input
     * @param pattern
     * @return true if matches, false if not
     */
    public static boolean matchPattern(String input, String pattern) {
        if (pattern == null) {
            return true;
        }

        boolean regexexpression = false;
        String regexexpressionstring = null;

        if (pattern.contains("%")) {
            regexexpression = true;

            if (regexexpressionstring == null) {
                regexexpressionstring = pattern.replace("%", ".*");
            }
        }
        if (pattern.contains("_")) {
            regexexpression = true;

            if (regexexpressionstring == null) {
                regexexpressionstring = pattern.replace("_", ".");
            } else {
                regexexpressionstring = regexexpressionstring.replace("_", ".");
            }
        }
        if (regexexpression) {
            return input.matches(regexexpressionstring);
        } else {
            //return input.contains(pattern);
            return input.equals(pattern);
        }
    }

    /**
     * Convert Bigquery field type to java.sql.Types
     *
     * @param columntype String of the Column type
     * @return java.sql.Types value of the given Columtype
     */
    public static int parseToSqlFieldType(String columntype) {
    	if (columntype.equals("FLOAT")) {
    		return java.sql.Types.DOUBLE;
    	} else if (columntype.equals("BOOLEAN")) {
    		return java.sql.Types.BOOLEAN;
    	} else if (columntype.equals("INTEGER")) {
    		return java.sql.Types.BIGINT;
    	} else if (columntype.equals("STRING")) {
    		return java.sql.Types.VARCHAR;
    	} else if (columntype.equals("BIGINT")){
    		return java.sql.Types.BIGINT;
    	} else if (columntype.equals("TIMESTAMP")){
    		return java.sql.Types.TIMESTAMP;
    	} else if (columntype.equals("DATE")){
    		return java.sql.Types.DATE;
    	} else if (columntype.equals("TIME")){
    		return java.sql.Types.TIME;
    	} else if (columntype.equals("DATETIME")){
    		return java.sql.Types.TIMESTAMP;
    	} else if (columntype.equals("BYTES")){
    		return java.sql.Types.BLOB;
    	} else {
    		return 0;
    	}        
    }

    public static Object getObjectOfType(String typeName, String value, int maxFieldSize) {
    	Object obj = getObjectOfType(typeName, value);
        if (maxFieldSize > 0 && maxFieldSize < Integer.MAX_VALUE && obj instanceof String && ((String) obj).length() > maxFieldSize)
        	return ((String) obj).substring(0, maxFieldSize);
        return obj;
    }

    public static Object getObjectOfType(String typeName, String value) {
    	int sqlType = parseToSqlFieldType(typeName);
    	return getObjectOfType(sqlType, value);
    }

    public static Object getObjectOfType(int sqlType, String value) {
    	switch (sqlType) {
    	case java.sql.Types.DOUBLE:
    		return Double.parseDouble(value);
    	case java.sql.Types.BOOLEAN:
    		return Boolean.parseBoolean(value);
    	case java.sql.Types.BIGINT:
    		return new BigInteger(value);
    	case java.sql.Types.TIMESTAMP:
			return new java.sql.Timestamp((long) (Double.parseDouble(value) * 1000));
    	case java.sql.Types.VARCHAR:
    	default:
    		return value;
    	}
    }
    public static int parseToNullType(String nulltype) {
        if ("REQUIRED".equals(nulltype)) {
        	// no nulls
            return NULLTYPE_REQUIRED;
        } else if ("NULLABLE".equals(nulltype)) {
        	// nullable
        	return NULLTYPE_NULLABLE;
        } else if ("REPEATED".equals(nulltype)) {
        	// repeated
        	return NULLTYPE_REPEATED;
        } else
        	return NULLTYPE_REQUIRED;
    }

	public static String parseToIsoNullType(String nulltype) {
		int nulltype0 = parseToNullType(nulltype);
		switch (nulltype0) {
			case 0:
	        	// no nulls
	            return "NO";
			case 1:
	        	// nullable
	            return "YES";
			case 2:
			default:
	        	// unknown
	            return "";
		}
	}

	/**
     * Reads Properties File from location
     *
     * @param filePath Location of the Properties File
     * @return Properties The Properties object made from the Properties File
     * @throws IOException if the load from file fails
     */
    public static Properties readFromPropFile(String filePath)
            throws IOException {
        // Read properties file.
        Properties properties = new Properties();
        properties.load(new FileInputStream(filePath));

        return properties;
    }

    /**
     * Starts a new query in async mode.
     *
     * @param bigquery  The bigquery instance, which is authorized
     * @param projectId The project's ID
     * @param querySql  The sql query which we want to run
     * @return A JobReference which we'll use to poll the bigquery, for its
     *         state, then for its mined data.
     * @throws IOException <p>
     *             if the request for initializing or executing job fails
     *             </p>
     */
    public static Job startQuery(Bigquery bigquery, String projectId,
            String querySql, TableReference tableRef) throws IOException {
        BQSupportFuncts.logger.info("Inserting Query Job: " + querySql.replace("\t","").replace("\n", " ").replace("\r", ""));
        projectId = BQSupportFuncts.decodeId(projectId);
        Job job = new Job();
        JobConfiguration config = new JobConfiguration();
        JobConfigurationQuery queryConfig = new JobConfigurationQuery();
        config.setQuery(queryConfig);

        job.setConfiguration(config);
        queryConfig.setQuery(querySql);
        if (tableRef != null) {
        	queryConfig.setAllowLargeResults(true);
        	queryConfig.setDestinationTable(tableRef);
        }
        
        Insert insert = bigquery.jobs().insert(querySql, job);
        insert.setProjectId(projectId);
        return insert.execute();
    }

    public static String decodeId(String projectId) {
    	return projectId.replace("__", ":").replace("_", ".");
    }

    public static String encodeId(String projectId) {
    	return projectId.replace(":" , "__").replace(".", "_");
    }

    public static TableReference extractDestTable(BQQueryParser parser, String projectId) {
    	final BQConnection conn = (BQConnection) parser.getConnection();
		if (!conn.getAllowLargeResults())
    		return null;
		String datasetId = conn.getLargeResultsDefaultDataset();
    	TableReference tableRef = null;
    	if (datasetId != null) {
    		tableRef = new TableReference();
    		tableRef.setDatasetId(datasetId);
    	} else
    		tableRef = parser.getTableRef();
       	if (tableRef != null) {
    		tableRef.setProjectId(projectId);
			byte[] bytesOfMessage = (parser.getQueryToParse() + String.valueOf(System.currentTimeMillis())).getBytes(StandardCharsets.UTF_8);
			try {
				MessageDigest md = MessageDigest.getInstance("MD5");
				byte[] thedigest = md.digest(bytesOfMessage);
				tableRef.setTableId("dv_temp_" + new SimpleDateFormat("yyyyMMddHHmmss").format(new Date()) + "_" + (new HexBinaryAdapter()).marshal(thedigest));
			} catch (NoSuchAlgorithmException e) {
				e.printStackTrace();
				return null;
			}
       	}
    	return tableRef;
    }

    public static void dropReferencedTable(Bigquery bigquery, TableReference tableRef) throws BQSQLException {
    	try {
    		bigquery.tables().delete(BQSupportFuncts.decodeId(tableRef.getProjectId()), tableRef.getDatasetId(), tableRef.getTableId()).execute();
    	} catch (IOException e) {
    		throw new BQSQLException("Something went wrong with the delete table call: " + tableRef, e);
    	}
    }

    public static String unquoteName(String name) {
    	return name.replaceAll("^\"|\"$", "").replace("\\\"", "\"").replace("\\\\", "\\");
    }

    public static String unquoteLiteral(String name) {
    	return name.replaceAll("^'|'$", "").replace("\\'", "'").replace("\\\\", "\\");
    }

	private static final String BRACE_QUOTED_REGEX =
			"\\[(?:(?:(?:(?:\\\")?([^:\\.\\\"]+)(?:\\\")?(?:[:\\.])))?(?:(?:\\\")?([^:\\.\\\"]+)(?:\\\")?(?:[:\\.])))?(?:\\\")?([^:\\.\\\"]+)(?:\\\")?\\]";
	public static void parseBraceQuoted(SourceTable sourceTable, String text) {
		Matcher mtch = Pattern.compile(BRACE_QUOTED_REGEX, Pattern.CASE_INSENSITIVE).matcher(text);
		if (mtch.find() && mtch.groupCount() == 3) {
			String grp = mtch.group(1);
			if (grp != null)
				sourceTable.setProject(grp);
			grp = mtch.group(2);
			if (grp != null)
				sourceTable.setDataset(grp);
			grp = mtch.group(3);
			if (grp != null)
				sourceTable.setName(grp);
		}
	}

    static final SimpleDateFormat format1ts = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
    static final SimpleDateFormat format2ts = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS");
    private static final SimpleDateFormat format1t = new SimpleDateFormat("HH:mm:ss.SSS");
	/** Parse date/time types */
	public static Date toTimestamp(String value, Calendar cal) throws SQLException {
	    try {
		    return java.sql.Time.valueOf(value);
	    } catch (IllegalArgumentException e) {
	    }
    	try {
    		if (Pattern.compile("\\d\\d:\\d\\d:\\d\\d.\\d{6}").matcher(value).matches())
    			value = value.substring(0, 12);
    		return new java.sql.Time(trimMillis(format1t.parse(value).getTime()));
    	} catch (ParseException e) { }
	    try {
		    return java.sql.Date.valueOf(value);
	    } catch (IllegalArgumentException e) { }
		try {
			return new Timestamp(trimMillis(format1ts.parse(value).getTime()));
		} catch (ParseException e) { }
		try {
			return new Timestamp(trimMillis(format2ts.parse(value).getTime()));
		} catch (ParseException e) { }
	    try {
	        long dbValue = new BigDecimal(value).movePointRight(3).longValue(); // movePointRight(3) =  *1000 (done before rounding) - from seconds (BigQuery specifications) to milliseconds (required by java). Precision under millisecond is discarded (BigQuery supports micro-seconds)
	        if (cal == null) {
	        	cal = Calendar.getInstance(TimeZone.getTimeZone("UTC")); // The time zone of the server that host the JVM should NOT impact the results. Use UTC calendar instead (which wont apply any correction, whatever the time zone of the data)
	        }
	        GregorianCalendar dbCal = (GregorianCalendar) GregorianCalendar.getInstance(TimeZone.getTimeZone("UTC"));
	        dbCal.setGregorianChange(new Date(Long.MIN_VALUE));
	        dbCal.setTimeInMillis(dbValue);
	        cal.set(Calendar.YEAR, dbCal.get(Calendar.YEAR));
	        cal.set(Calendar.MONTH, dbCal.get(Calendar.MONTH));
	        cal.set(Calendar.DAY_OF_MONTH, dbCal.get(Calendar.DAY_OF_MONTH));
	        cal.set(Calendar.HOUR_OF_DAY, dbCal.get(Calendar.HOUR_OF_DAY));
	        cal.set(Calendar.MINUTE, dbCal.get(Calendar.MINUTE));
	        cal.set(Calendar.SECOND, dbCal.get(Calendar.SECOND));
	        cal.set(Calendar.MILLISECOND, 0);
	        return new Timestamp(cal.getTimeInMillis());
	    }
	    catch (NumberFormatException e) {
	        throw new BQSQLException(e);
	    }
	}

	private static long trimMillis(long time) {
        GregorianCalendar cal = (GregorianCalendar) GregorianCalendar.getInstance(TimeZone.getTimeZone("UTC"));
        cal.setGregorianChange(new Date(Long.MIN_VALUE));
        cal.setTimeInMillis(time);
        cal.set(Calendar.MILLISECOND, 0);
		return cal.getTimeInMillis();
	}
}
