/**
 * Copyright (c) 2015, STARSCHEMA LTD.
 * All rights reserved.

 * Redistribution and use in source and binary forms, with or without modification, are permitted
 * provided that the following conditions are met:

 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.

 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Modified by
 * @author  Data Virtuality GmbH
 * @date    $Date$
 */
package net.starschema.clouddb.jdbc.list;

import java.util.ArrayList;
import java.util.List;

import net.starschema.clouddb.jdbc.JdbcGrammarParser;
import net.starschema.clouddb.jdbc.antlr.sqlparse.ColumnCallException;
import net.starschema.clouddb.jdbc.antlr.sqlparse.TreeParsingException;

import org.antlr.runtime.tree.Tree;

/**
 * This class extends the basic Node with the following attributes
 * <li> SelectStatement - which contains the SubQuery
 * <li>alias - for the SelectStatement
 *
 * @author Attila Horvath, Balazs Gunics
 */
public class SubQuery extends Node {
    String alias = null;
    SelectStatement selectStatement = null;

    TreeBuilder builder;

    /** Is this SubQuery part of a join? */
    private boolean isPartOfJoin = false;

    private String uniqueId;
    // EXTENSION
    // ISSUE: DVCORE-2462
    // FIXED BY: Roman Pedchenko
    // DATE: 2014.11.19

    private boolean isGenerated = false;

    // END EXTENSION
    /** Setter for isPartOfJoin, setting it true means that this subquery is part of a join */
    public void setisPartOfJoin() {
        isPartOfJoin = true;
    }

    /** Getter for isPartOfJoin() */
    public boolean isPartOfJoin() {
        return isPartOfJoin;
    }
    
	// EXTENSION
	// ISSUE: DVCORE-2462
	// FIXED BY: Roman Pedchenko
	// DATE: 2014.11.19
	// /**
    public boolean isGenerated() {
		return isGenerated;
	}

	public void setGenerated(boolean isGenerated) {
		this.isGenerated = isGenerated;
	}

	/**
	// END EXTENSION
     * Constructor for SubQuery, which builds a SubQuery from a selectStatement
     * with the given alias
     * @param alias - the SubQuerys name
     * @param builder - the TreeBuilder for the helper functions
     * @param selectStatement - to be contained in the SubQuery
     */
    public SubQuery(String alias, TreeBuilder builder,
            SelectStatement selectStatement) {
        this.builder = builder;
        this.selectStatement = selectStatement;
        selectStatement.setParent(this);
        this.alias = alias;
        this.tokenType = JdbcGrammarParser.SUBQUERY;
        this.tokenName = JdbcGrammarParser.tokenNames[JdbcGrammarParser.SUBQUERY];
        this.uniqueId = builder.getuniqueid();
    }

    /**
     * Constructor for SubQuery, which builds a SubQuery from a selectStatement
     * with the given alias and with the given uniqueId
     * @param alias - the SubQuerys name
     * @param builder - the TreeBuilder for the helper functions
     * @param selectStatement - to be contained in the SubQuery
     * @param uniqueId - the uniqueId for the subquery
     */
    public SubQuery(String alias, TreeBuilder builder,
            SelectStatement selectStatement, String uniqueId) {
        this.builder = builder;
        this.selectStatement = selectStatement;
        selectStatement.setParent(this);
        this.alias = alias;
        this.tokenType = JdbcGrammarParser.SUBQUERY;
        this.tokenName = JdbcGrammarParser.tokenNames[JdbcGrammarParser.SUBQUERY];
        this.uniqueId = uniqueId;
    }

    /**
     * Constructor for SubQuery, which builds a SubQuery from an ANTLR tree
     * @param t - the ANTLR tree
     * @param treeBuilder - the TreeBuilder for the helper functions
     * @throws TreeParsingException - if we fail to parse the ANTLR tree
     */
    public SubQuery(Tree t, TreeBuilder treeBuilder) throws TreeParsingException {
        this.builder = treeBuilder;
        this.uniqueId = this.builder.getuniqueid();
        this.build(t, this.builder);
    }

    /**
     * Builder to parse out the ANTLR tree
     * @param t - the ANTLR tree
     * @param builder - the TreeBuilder for the helper functions
     * @throws TreeParsingException - if we fail to parse the ANTLR tree
     */
    public void build(Tree t, TreeBuilder builder) throws TreeParsingException {
        if (t.getType() == JdbcGrammarParser.SUBQUERY) {

            this.tokenName = JdbcGrammarParser.tokenNames[t.getType()];
            this.logger.debug("BUILDING " + this.tokenName);
            this.tokenType = t.getType();
            for (int i = 0; i < t.getChildCount(); i++) {
                Tree child = t.getChild(i);
                switch (child.getType()) {
                    case JdbcGrammarParser.SELECTSTATEMENT:
                        try {
                            this.selectStatement = new SelectStatement(child,
                                    builder);
                        } catch (ColumnCallException e) {
                            throw new TreeParsingException(e);
                        }
                        selectStatement.parent = this;
                        break;
                    case JdbcGrammarParser.ALIAS:
                        this.alias = child.getChild(0).getText();
                    default:
                        break;
                }
            }
            if (this.getAlias() == null) {
                // we give alias as unique id
                this.alias = this.uniqueId;
            }
        } else {
            throw new TreeParsingException("This Tree is not a SUBQUERY");
        }
    }

    /** Getter for the SuqbQuery alias */
    public String getAlias() {
        return this.alias;
    }

    /**
     * Iterates through the Columns and functions of this subquery and
     * puts their Synonyms into a SynonymContainer, then returns these
     * SynonymContainers as a List
     * @return - the Synonyms of the Columns/Functions
     */
    public List<SynonymContainer> getAvailableResources() {
        Expression expression = this.selectStatement.getExpression();
        List<ColumnCall> columns = expression.getColumns();
        List<FunctionCall> functionCalls = expression.getFunctionCalls();
	    // EXTENSION
	    // ISSUE: DVCORE-3121
	    // FIXED BY: Dmitry Pogorelov
	    // DATE: 2015.09.29
	    List<FunctionParameter> singleExpressions = expression.getSingleExpressions();
        //END EXTENSION

        List<SynonymContainer> container = new ArrayList<SynonymContainer>();

        if (columns != null) {
            for (ColumnCall columnCall : columns) {

                SynonymContainer columnCallSynonymContainer = new SynonymContainer(
                        columnCall);

                List<String> synonyms = columnCall.getSynonyms();
                if (synonyms != null) {
                    for (String string : synonyms) {
                        columnCallSynonymContainer.addSynonym(string);
                    }
                }
                container.add(columnCallSynonymContainer);
            }
        }
        if (functionCalls != null) {
            for (FunctionCall functionCall : functionCalls) {
                SynonymContainer columnCallSynonymContainer = new SynonymContainer(
                        functionCall);
                List<String> synonyms = functionCall.getSynonyms();
                if (synonyms != null) {
                    for (String string : synonyms) {
                        columnCallSynonymContainer.addSynonym(string);
                    }
                }
                container.add(columnCallSynonymContainer);
            }
        }
	    // EXTENSION
	    // ISSUE: DVCORE-3121
	    // FIXED BY: Dmitry Pogorelov
	    // DATE: 2015.09.29
	    if (singleExpressions != null) {
		    for (FunctionParameter singleExpression : singleExpressions) {
			    SynonymContainer columnCallSynonymContainer = new SynonymContainer(
					    singleExpression);
			    List<String> synonyms = singleExpression.getSynonyms();
			    if(synonyms!=null){
				    for (String string : synonyms) {
					    columnCallSynonymContainer.addSynonym(string);
				    }
			    }
			    container.add(columnCallSynonymContainer);
		    }
	    }
	    // END EXTENSION
        return container;
    }

    /** Getter for the SelectStatement */
    public SelectStatement getSelectStatement() {
        return this.selectStatement;
    }

    /** Getter for the SubQuerys uniqueId */
    public String getUniqueId() {
        return this.uniqueId;
    }

    @Override
    public String toPrettyString() {
        return this.toPrettyString(-1);
    }

    @Override
    public String toPrettyString(int level) {
        String result;
        int newlevel = level < 0 ? -1 : level;
        if (this.alias != null) {
            result = this.tab(level - 1) + "("
                    + this.selectStatement.toPrettyString(newlevel) + ") "
                    + this.alias + newline;
        } else {
            result = this.tab(level - 1) + "("
                    + this.selectStatement.toPrettyString(newlevel) + ")"
                    + newline;
        }
        return result;
    }
}
